from pGerrit.restAPIwrapper import GerritRest
from pGerrit.client import GerritClient
from pGerrit.utils import urljoin, urlformat
from pGerrit.queryMeta import QueryMeta

class GerritChange(GerritClient, metaclass=QueryMeta):
    """Interface to the Gerrit REST API.
    :arg str url: The full URL to the server, including the `http(s)://`
        prefix.
    :arg auth: (optional) Authentication handler.  Must be derived from
        `requests.auth.HTTPDigestAuth`.
    :arg boolean verify: (optional) Set to False to disable verification of
        SSL certificates.
    :arg requests.adapters.BaseAdapter adapter: (optional) Custom connection
        adapter. See
        https://requests.readthedocs.io/en/master/api/#requests.adapters.BaseAdapter
    """
    _endpoint = "/a/changes/{}"
    _args = ["id"]

    def __init__(self, host, gerritID=None, auth=None, verify=True, adapter=None):
        """See class docstring."""
        super().__init__(host, auth=auth, verify=verify, adapter=adapter)
        self.id = gerritID

    @GerritRest.get()
    def query(self, *args, **kwargs):
        return urljoin(self.host, urlformat(GerritChange._endpoint, ""))

    @GerritRest.get()
    def info(self, *args, **kwargs):
        return urljoin(self.host, urlformat(GerritChange._endpoint, self.id))

    def revision(self, revisionID):
        return GerritChangeRevision(self.host, self.id, revisionID, **self.kwargs)

    def current_revision(self):
        return GerritChangeRevision(self.host, self.id, "current", **self.kwargs)

    def is_merge(self):
        revision = self.current_revision()
        if len(revision.commit().parents) == 2:
            return True
        else:
            return False

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def detail(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def topic(self, *args, **kwargs):
        pass

    @GerritRest.put
    @GerritRest.url_wrapper("topic")
    def set_topic(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def submitted_together(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper("in")
    def _in(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def comments(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def robotcomments(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def drafts(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def check(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def edit(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def reviewers(self, *args, **kwargs):
        pass

    @GerritRest.post
    @GerritRest.url_wrapper("reviewers")
    def add_reviewer(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def suggest_reviewers(self, *args, **kwargs):
        pass

    @GerritRest.post
    @GerritRest.url_wrapper()
    def rebase(self, payload=None, headers=None):
        pass

    @GerritRest.post
    @GerritRest.url_wrapper("edit:publish")
    def edit_publish(self, payload=None, headers=None):
        pass

    @GerritRest.post
    @GerritRest.url_wrapper("edit")
    def edit_restore(self, payload=None, headers=None):
        pass

    @GerritRest.delete
    @GerritRest.url_wrapper("edit")
    def edit_delete(self, payload=None, headers=None):
        pass

    @GerritRest.delete
    @GerritRest.url_wrapper("")
    def delete_change(self, payload=None, headers=None):
        pass

class GerritChangeRevision(GerritChange):
    """Interface to the Gerrit REST API.
    :arg str url: The full URL to the server, including the `http(s)://`
        prefix. If `auth` is given, `url` will be automatically adjusted to
        include Gerrit's authentication suffix.
    :arg auth: (optional) Authentication handler.  Must be derived from
        `requests.auth.HTTPDigestAuth`.
    :arg boolean verify: (optional) Set to False to disable verification of
        SSL certificates.
    :arg requests.adapters.BaseAdapter adapter: (optional) Custom connection
        adapter. See
        https://requests.readthedocs.io/en/master/api/#requests.adapters.BaseAdapter
    """
    _endpoint = "/a/changes/{}/revisions/{}"
    _args = ["id", "revisionID"]

    def __init__(self, host, gerritID, revisionID, auth=None, verify=True, adapter=None):
        """See class docstring."""
        super().__init__(host, gerritID, auth=auth, verify=verify, adapter=adapter)
        self.revisionID = revisionID

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def commit(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def actions(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def review(self, *args, **kwargs):
        pass

    @GerritRest.post
    @GerritRest.url_wrapper("review")
    def set_review(self, payload=None, headers=None):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def related(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def patch(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def mergeable(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def submit_type(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def drafts(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def comments(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def files(self, *args, **kwargs):
        pass

    @GerritRest.post
    @GerritRest.url_wrapper()
    def cherrypick(self, payload=None, headers=None):
        pass

    def file(self, fileID):
        return GerritChangeRevisionFile(self.host, self.id, self.revisionID, fileID, **self.kwargs)

    def reviwer(self, accountID):
        return GerritChangeRevisionReviewer(self.host, self.id, self.revisionID, accountID, **self.kwargs)

class GerritChangeRevisionReviewer(GerritChangeRevision):
    """Interface to the Gerrit REST API.
    :arg str url: The full URL to the server, including the `http(s)://`
        prefix. If `auth` is given, `url` will be automatically adjusted to
        include Gerrit's authentication suffix.
    :arg auth: (optional) Authentication handler.  Must be derived from
        `requests.auth.HTTPDigestAuth`.
    :arg boolean verify: (optional) Set to False to disable verification of
        SSL certificates.
    :arg requests.adapters.BaseAdapter adapter: (optional) Custom connection
        adapter. See
        https://requests.readthedocs.io/en/master/api/#requests.adapters.BaseAdapter
    """
    _endpoint = "/a/changes/{}/revisions/{}/reviewers/{}"
    _args = ["id", "accountID", "revisionID", "accountID"]

    def __init__(self, host, gerritID, revisionID, accountID, auth=None, verify=True, adapter=None):
        """See class docstring."""
        super().__init__(host, gerritID, revisionID, auth=auth, verify=verify, adapter=adapter)
        self.accountID = accountID

    @GerritRest.get()
    def list(self, *args, **kwargs):
        return urljoin(self.host, urlformat(self._endpoint, self.id, self.revisionID, ""))

    @GerritRest.delete
    def delete_vote(self, label, *args, **kwargs):
        return urljoin(self.host, urlformat(self._endpoint, self.id, self.revisionID, self.accountID), "votes", label)


class GerritChangeRevisionFile(GerritChangeRevision):
    """docstring for GerritChangeRevisionFile"""
    _endpoint = "/a/changes/{}/revisions/{}/files/{}"
    _args = ["id", "revisionID", "fileID"]

    def __init__(self, host, gerritID, revisionID, fileID, auth=None, verify=True, adapter=None):
        super(GerritChangeRevisionFile, self).__init__(host, gerritID, revisionID, auth=auth, verify=verify, adapter=adapter)
        self.fileID = fileID

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def content(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def diff(self, *args, **kwargs):
        pass

    @GerritRest.get(raw=True)
    @GerritRest.url_wrapper()
    def download(self, *args, **kwargs):
        pass

    @GerritRest.get()
    @GerritRest.url_wrapper()
    def blame(self, *args, **kwargs):
        pass

    def is_binary(self):
        file_info = getattr(self.files(), self.fileID)
        if hasattr(file_info, "binary") and file_info.binary == True:
            return True
        else:
            return False

    @GerritRest.get()
    def get_history_log(self, commit=None, *args, **kwargs):
        project = self.info().project
        commit = commit or self.commit().commit
        return urljoin(self.host, "a/plugins", "gitiles", project, "+log", commit, self.fileID)

    @GerritRest.put
    def edit(self, payload, headers=None):
        return urljoin(self.host, "/a/changes/", self.id, "/edit/", urlformat("{}", self.fileID))

    @GerritRest.get()
    def edit_retrieve(self, headers=None):
        return urljoin(self.host, "/a/changes/", self.id, "/edit/", urlformat("{}", self.fileID))




from urllib.parse import quote
import re

def urljoin(*args):
    return "/".join(map(lambda x: str(x).rstrip('/').lstrip('/'), args)) + ("/" if args[len(args) - 1].endswith("/") else "")

def urlformat(formattedString, *args):
    args = [quote(str(arg), safe="") for arg in args]
    return formattedString.format(*args)

def parseCookieFile(cookiefile):
    """Parse a cookies.txt file and return a dictionary of key value pairs
    compatible with requests."""

    cookies = {}
    with open (cookiefile, 'r') as fp:
        for line in fp:
            if not re.match(r'^\#', line):
                lineFields = line.strip().split('\t')
                cookies[lineFields[5]] = lineFields[6]
    return cookies
        

# def simpleDictToObject(dict, Class):
#     


import sys
import os
import requests

import unittest
from pGerrit.client import GerritClient
from pGerrit.change import GerritChange, GerritChangeRevision, GerritChangeRevisionFile
from pGerrit.utils import urljoin
from requests.auth import HTTPBasicAuth
from pGerrit.utils import urljoin
from types import SimpleNamespace

import requests_cache
requests_cache.disabled()

# To run this test, User need to set environment variable below, the change should match following status in order to pass all tests
# 1. change should be in master branch
# 2. change is already merged
# 3. change has a topic
# 4. you can set a topic on this change
# 5. this change must have at least one reviewer
# 6. you can set label Code-review on this change +1

# export GERRIT_USERNAME=chengyang # use your own username
# export GERRIT_PASSWORD=pFFPgP/giebge0xxxxxxxxxxxxxxxxxx  # This is retreived from Gerrit->Settings->HTTP Credentials
# export GERRIT_HOST=https://android-review.googlesource.com/
# export GERRIT_CHANGE_NUMBER=1234567  #find a change that meets requirement above

# then run python -m unittest pGerrit/tests/test.py

g_username = os.environ.get("GERRIT_USERNAME")
assert g_username is not None

g_password = os.environ.get("GERRIT_PASSWORD")
assert g_password is not None

g_host =  os.environ.get("GERRIT_HOST")
assert g_host is not None

g_id = os.environ.get("GERRIT_CHANGE_NUMBER")
assert g_id is not None


class TestChange(unittest.TestCase):
    """docstring for TestGerrit"""
    def setUp(self):
        requests.packages.urllib3.disable_warnings()
        self.auth = HTTPBasicAuth(g_username, g_password)
        self.client = GerritClient(g_host, auth=self.auth, verify=False, cache=False)
        self.change = self.client.change(g_id)

    def tearDown(self):
        pass

    def testChangeDetail(self):
        detail = self.change.detail(o=["ALL_COMMITS", "CURRENT_REVISION"])
        self.assertIsInstance(detail, SimpleNamespace)
        self.assertEqual(detail.branch, "master")

    def testChangeInfo(self):
        info = self.change.info(o=["ALL_COMMITS", "CURRENT_REVISION"])
        self.assertIsInstance(info, SimpleNamespace)
        self.assertEqual(info.branch, "master")

    def testChangeIsMerge(self):
        self.assertTrue(self.change.is_merge())

    def testChangeCurrentRevision(self):
        revision = self.change.current_revision()
        self.assertIsInstance(revision, GerritChangeRevision)

    def testChangeRevision(self):
        revision = self.change.revision("current")
        self.assertIsInstance(revision, GerritChangeRevision)

    def testChangeQuery(self):
        results = self.client.change.query(q=g_id)
        self.assertIsInstance(results, list)
        for result in results:
            self.assertIsInstance(result, SimpleNamespace)

    def testChangeTopic(self):
        topic = self.change.topic()
        self.assertIsInstance(topic, str)

    def testSetChangeTopic(self):
        target_topic = "test_set_topic"
        self.change.set_topic(target_topic)
        topic = self.change.topic()
        self.assertEqual(topic, target_topic)

    def testChangeSubmitted_together(self):
        submitted_together = self.change.submitted_together()
        self.assertIsInstance(submitted_together, list)

    def testChange_in(self):
        _in = self.change._in()
        self.assertIsInstance(_in, SimpleNamespace)

    def testChangeComments(self):
        comments = self.change.comments()
        self.assertIsInstance(comments, SimpleNamespace)

    def testChangeDrafts(self):
        drafts = self.change.drafts()
        self.assertIsInstance(drafts, SimpleNamespace)

    def testChangeCheck(self):
        check = self.change.check()
        self.assertIsInstance(check, SimpleNamespace)

    @unittest.skip("This change cannot be rebased, so skip")
    def testChangeRebase(self):
        rebase = self.change.rebase()
        # self.assertIsInstance(edit, SimpleNamespace)

    @unittest.skip("Edit will return empty unless the change is in the edit status")
    def testChangeEdit(self):
        edit = self.change.edit()
        self.assertIsInstance(edit, SimpleNamespace)

    @unittest.skip("Edit will return empty unless the change is in the edit status")
    def testChangeEditPublish(self):
        edit = self.change.edit_publish()
        self.assertIsInstance(edit, SimpleNamespace)

    @unittest.skip("Edit will return empty unless the change is in the edit status")
    def testChangeEditDelete(self):
        edit = self.change.edit_delete()
        self.assertIsInstance(edit, SimpleNamespace)

    def testChangeReviewers(self):
        reviewers = self.change.reviewers()
        self.assertIsInstance(reviewers[0], SimpleNamespace)

    def testChangeAddReviewers(self):
        payload = {"reviewer":g_username}
        res = self.change.add_reviewer(payload=payload)
        self.assertEqual(res.status_code, 200)

    def testChangeSuggest_reviewers(self):
        suggest_reviewers = self.change.suggest_reviewers()
        self.assertIsInstance(suggest_reviewers[0], SimpleNamespace)

class TestChangeRevisionReviewer(unittest.TestCase):
    """docstring for Test"""
    def setUp(self):
        requests.packages.urllib3.disable_warnings()
        self.auth = HTTPBasicAuth(g_username, g_password)
        self.client = GerritClient(g_host, auth=self.auth, verify=False)
        self.change = self.client.change(g_id)
        self.revision = self.change.revision("1")
        self.current = self.change.current_revision()
        self.reviewer = self.current.reviwer("")

    def tearDown(self):
        pass

    def testRevisionReviwerList(self):
        files = self.reviewer.list()
        self.assertIsInstance(files, list)

    @unittest.skip("Delete vote can only delete others vote, so cannot be idempotent")
    def testRevisionReviwerDeleteVote(self):
        files = self.reviewer.delete_vote()
        self.assertIsInstance(files, SimpleNamespace)

class TestChangeRevision(unittest.TestCase):
    """docstring for Test"""
    def setUp(self):
        requests.packages.urllib3.disable_warnings()
        self.auth = HTTPBasicAuth(g_username, g_password)
        self.client = GerritClient(g_host, auth=self.auth, verify=False)
        self.change = self.client.change(g_id)
        self.revision = self.change.revision("1")
        self.current = self.change.current_revision()

    def tearDown(self):
        pass

    def testRevisionFiles(self):
        files = self.revision.files()
        self.assertIsInstance(files, SimpleNamespace)

    def testRevisionCommit(self):
        commit = self.revision.commit()
        self.assertIsInstance(commit, SimpleNamespace)

    def testRevisionActions(self):
        actions = self.revision.actions()
        self.assertIsInstance(actions, SimpleNamespace)

    def testRevisionReview(self):
        review = self.revision.review()
        self.assertIsInstance(review, SimpleNamespace)

    def testRevisionRelated(self):
        related = self.revision.related()
        self.assertIsInstance(related, SimpleNamespace)

    # this change has 2 parents, so patch interface will raise error
    @unittest.expectedFailure
    def testRevisionPatch(self):
        patch = self.revision.patch()
        self.assertIsInstance(patch, SimpleNamespace)

    # this change is abandoned, so mergeable interface will rasie error
    # @unittest.expectedFailure
    @unittest.skip
    def testRevisionMergeable(self):
        mergeable = self.revision.mergeable()
        self.assertIsInstance(mergeable, SimpleNamespace)

    def testRevisionSubmit_type(self):
        submit_type = self.revision.submit_type()
        self.assertIsInstance(submit_type, str)

    def testRevisionDrafts(self):
        drafts = self.revision.drafts()
        self.assertIsInstance(drafts, SimpleNamespace)

    def testRevisionComments(self):
        comments = self.revision.comments()
        self.assertIsInstance(comments, SimpleNamespace)

    @unittest.skip
    def testRevisionSetReview(self):
        info = {"labels":{"Code-Review":1}}
        res = self.revision.set_review(payload=info)
        info["labels"]["Code-Review"] = 0
        res = self.revision.set_review(payload=info)
        self.assertEqual(res.status_code, 200)

class TestRevisionFile(unittest.TestCase):
    """docstring for TestRevisionFile"""
    """docstring for Test"""
    def setUp(self):
        requests.packages.urllib3.disable_warnings()
        self.auth = HTTPBasicAuth(g_username, g_password)
        self.client = GerritClient(g_host, auth=self.auth, verify=False)
        self.change = self.client.change(g_id)
        self.current = self.change.current_revision()
        self.file = self.current.file("1")

    def tearDown(self):
        pass

    def testRevisionFileContent(self):
        content = self.file.content()
        self.assertIsInstance(content, str)

    def testRevisionFileDiff(self):
        diff = self.file.diff()
        self.assertIsInstance(diff, SimpleNamespace)

    # this interface will return binary, so json decoder cannot handle it perfectly
    @unittest.skip
    def testRevisionFileDownload(self):
        download = self.file.download()
        self.assertIsInstance(download, SimpleNamespace)

    def testRevisionFileBlame(self):
        blame = self.file.blame()
        self.assertIsInstance(blame[0], SimpleNamespace)

    def testRevisionFileIsBinary(self):
        self.assertFalse(self.file.is_binary())

    def testRevisionGetHistoryLog(self):
        logs = self.file.get_history_log(format="JSON")
        for log in logs.log:
            self.assertIsInstance(log, SimpleNamespace)

class TestAccess(unittest.TestCase):
    """docstring for TestGerrit"""
    def setUp(self):
        requests.packages.urllib3.disable_warnings()
        self.auth = HTTPBasicAuth(g_username, g_password)
        self.client = GerritClient(g_host, auth=self.auth, verify=False)

    def tearDown(self):
        pass

    def testAccessQuery(self):
        access = self.client.access.query(project="All-Projects")
        self.assertIsInstance(access, SimpleNamespace)




import requests
from requests.packages.urllib3.util import Retry
from requests.adapters import HTTPAdapter
import requests_cache

class GerritClient(object):
    """Interface to the Gerrit REST API.
    :arg str url: The full URL to the server, including the `http(s)://`
        prefix. If `auth` is given, `url` will be automatically adjusted to
        include Gerrit's authentication suffix.
    :arg auth: (optional) Authentication handler.  Must be derived from
        `requests.auth.AuthBase`.
    :arg boolean verify: (optional) Set to False to disable verification of
        SSL certificates.
    :arg requests.adapters.BaseAdapter adapter: (optional) Custom connection
        adapter. See
        https://requests.readthedocs.io/en/master/api/#requests.adapters.BaseAdapter
    """

    def __init__(self, host, auth=None, cookies=None, verify=True, adapter=None, cache=True, cache_expire=3):
        """See class docstring."""
        self.host = host
        if cache:
            self.session = requests_cache.CachedSession(expire_after=cache_expire)
        else:
            self.session = requests.session()

        self.verify = verify
        if not adapter:
            retry = Retry(
                total=5,
                read=5,
                connect=5,
                backoff_factor=0.3,
                status_forcelist=(500, 502, 504),
            )
            adapter = HTTPAdapter(max_retries=retry)
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)
        self.adapter = adapter

        if auth:
            self.session.auth = auth
        elif cookies:
            self.session.cookies = cookies

        self.kwargs = {"auth": auth, "verify": verify, "adapter":adapter}

        if not self.host.endswith("/"):
            self.host += "/"

    def __del__(self):
        self.session.close()

    @property
    def change(self):
        from pGerrit.change import GerritChange
        return GerritChange(self.host, gerritID=None, auth=self.session.auth, verify=self.verify, adapter=self.adapter)

    @property
    def access(self):
        from pGerrit.Access import GerritAccess
        return GerritAccess(self.host, auth=self.session.auth, verify=self.verify, adapter=self.adapter)


#This file is for implementiing GerritClient.change.query and GerritClient.change(xxxxx).

from pGerrit.client import GerritClient

class QueryMeta(type):
    def __new__(cls, name, bases, attrs):
        if GerritClient in bases:
            attrs['__call__'] = meta_call
        return super().__new__(cls, name, bases, attrs)

def meta_call(self, *args, **kwargs):
    if len(args) == 0:
        return self
    else:
        return self.__class__(self.host, gerritID=args[0] if len(args) > 0 else kwargs["gerritID"], auth=self.session.auth, verify=self.verify, adapter=self.adapter)


from pGerrit.restAPIwrapper import GerritRest
from pGerrit.client import GerritClient
from pGerrit.utils import urljoin, urlformat
from pGerrit.queryMeta import QueryMeta

class GerritAccess(GerritClient, metaclass=QueryMeta):
    """Interface to the Gerrit REST API.
    :arg str url: The full URL to the server, including the `http(s)://`
        prefix.
    :arg auth: (optional) Authentication handler.  Must be derived from
        `requests.auth.HTTPDigestAuth`.
    :arg boolean verify: (optional) Set to False to disable verification of
        SSL certificates.
    :arg requests.adapters.BaseAdapter adapter: (optional) Custom connection
        adapter. See
        https://requests.readthedocs.io/en/master/api/#requests.adapters.BaseAdapter
    """
    _endpoint = "/a/access/{}"

    def __init__(self, host, auth=None, verify=True, adapter=None):
        """See class docstring."""
        super().__init__(host, auth=auth, verify=verify, adapter=adapter)

    @GerritRest.get()
    def query(self, *args, **kwargs):
        return urljoin(self.host, urlformat(GerritAccess._endpoint, ""))
